/*
 * VNPForth - Compiled native Forth for IA32 Linux systems
 * Copyright (C) 2001  Simon Baldwin (simonb@caldera.com)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <assert.h>
#include <sys/param.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <errno.h>

#include "pstore.h"
#include "ident.h"
#include "ifile.h"
#include "sfile.h"
#include "getopts.h"

#include "codegen.h"

/* Definitions. */
#define CELL_SIZE		4
#define STABTYPE_VOID		1
#define STABTYPE_CELL		2
#define STABTYPE_UCELL		3
#define STABTYPE_CHAR		4

/* True/False macros. */
#define FALSE			(0)
#define TRUE			(!FALSE)

/* External definitions. */
extern char		*argv0;

/* Sequential count of definitions encountered. */
static int		defcount	= 0;

/* Pointer to the symbol table entry of any currently entered definition. */
static struct symrec	*currdef	= NULL;

/* Sequential count of debugging and begin-end labels. */
static int		lp_seq		= 0;
static int		lm_seq		= 0;
static int		be_seq		= 0;
static int		sc_seq		= 0;

/* Table of register mappings, from register number to x86 register string. */
static const struct {
	const int	preserved;
	const char	*x86reg;
} registers[] = {
	{ FALSE, "%eax" },	/* Map R0 to %eax, not preserved over calls. */
	{ TRUE,  "%ebx" }	/* Map R1 to %ebx, preserved over calls. */
};
#define	NUM_REGISTERS	(int)(sizeof (registers) / sizeof (registers[0]))


/*
 * fatal()
 *
 * Fatal error in pcode.  We should not get to here; if we do, something
 * unutterably nasty has happened, and we'll dump core in the hope of getting
 * to it.
 */
static void
fatal (const char *format, const char *message)
{
	assert (format != NULL && message != NULL);
	fprintf (stderr, "%s: code generation error:\n", get_ifname ());
	fprintf (stderr, "    ");
	fprintf (stderr, format, message);
	fprintf (stderr, "\n");
	abort ();
}


/*
 * generate_preamble()
 *
 * X86 debugging preamble generation routine.
 */
static void
generate_preamble (const char *source_path)
{
	char	cwd[MAXPATHLEN];		/* Getcwd buffer */
	assert (source_path != NULL);

	/* Write common file preamble. */
	printf_sfile ("\t.file\t\"%s\"\n", source_path);
	printf_sfile ("\t.version\t\"01.01\"\n");
	printf_sfile ("forth_compiled.:\n");

	/* Add profiling if required. */
	if (get_pflag ())
		printf_sfile ("\tnop\n");

	/* Do no other preamble if no debug set. */
	if (get_gflag ())
	    {
		/* Add getcwd entry as part of source file location data. */
		if (getcwd (cwd, sizeof (cwd)) != NULL)
			printf_sfile (".stabs\t\"%s/\",100,0,0,Ltext0\n", cwd);

		/* Add information relating to the source file. */
		printf_sfile (".stabs\t\"%s\",100,0,0,Ltext0\n", source_path);
		printf_sfile ("\t.stabs\t\"forth_compiled.\",0x3c,0,0,0\n");

		/* Add details on the basic types Forth uses. */
		printf_sfile (".text\nLtext0:\n");
		printf_sfile (".stabs\t\"void:t%d=%d;\",128,0,0,0\n",
						STABTYPE_VOID, STABTYPE_VOID);
		printf_sfile (".stabs\t\"CELL:t%d=r%d;%d;%d;\",128,0,0,0\n",
						STABTYPE_CELL, STABTYPE_CELL,
						INT_MIN, INT_MAX);
		printf_sfile (".stabs\t\"UCELL:t%d=r%d;%d;%d;\",128,0,0,0\n",
						STABTYPE_UCELL, STABTYPE_CELL,
						0, UINT_MAX);
		printf_sfile (".stabs\t\"CHAR:t%d=r%d;%d;%d;\",128,0,0,0\n",
						STABTYPE_CHAR, STABTYPE_CHAR,
						0, CHAR_MAX);
	    }
}


/*
 * debug_variable()
 *
 * X86 variable debug symbols output.
 */
static void
debug_variable (struct symrec *symp)
{
	assert (symp != NULL);

	/* Ignore call if no debug set. */
	if (get_gflag ())
	    {
		/* Write out stabs intro. */
		printf_sfile (".stabs\t\"%s", symp->name);

		/* If single-cell, this is not an array, otherwise it is. */
		if (symp->units == P_CELL && symp->len == 1)
		    {
			/* Write simple single-cell variable stab. */
			printf_sfile (":G%d", STABTYPE_CELL);
		    }
		else
		    {
			/* Write array stab. */
			printf_sfile (":Gar%d;0;%d;%d", STABTYPE_CELL,
					symp->len - 1,
					(symp->units == P_CELL) ?
						STABTYPE_CELL : STABTYPE_CHAR);
		    }

		/* Write stab conclusion. */
		printf_sfile ("\",40,0,0,%s\n", symp->name);
	    }
}


/*
 * debug_statement()
 *
 * X86 source statement debug generation.
 */
static void
debug_statement (int linenum)
{
	/* Ignore call if no debug set. */
	if (get_gflag ())
	    {
		/* Write the source line symbol. */
		printf_sfile (".stabn 68,0,%d,.LM%d-%s\n", linenum, ++lm_seq,
				(currdef != NULL) ? currdef->name : "main");

		/* Write the accompanying symbol. */
		printf_sfile (".LM%d:\n", lm_seq);
	    }
}


/*
 * debug_function()
 * debug_function_begin()
 * debug_function_end()
 * debug_function_cleanup()
 *
 * X86 function debug generation.
 */
static void
debug_function (const char *name)
{
	assert (name != NULL);

	/* Ignore call if no debug set. */
	if (get_gflag ())
	    {
		/* Write out the function stab header. */
		printf_sfile (".stabs \"%s:F%d\",36,0,0,%s\n", name,
							STABTYPE_VOID, name);
	    }
}
static void
debug_function_begin (const char *name)
{
	assert (name != NULL);

	/* Ignore call if no debug set. */
	if (get_gflag ())
	    {
		/* Write out a begin label. */
		printf_sfile (".LBB%d:\n", ++be_seq);
	    }
}
static void
debug_function_end (const char *name)
{
	assert (name != NULL);

	/* Ignore call if no debug set. */
	if (get_gflag ())
	    {
		/* Write out an end label. */
		printf_sfile (".LBE%d:\n", be_seq);
	    }
}
static void
debug_function_cleanup (const char *name)
{
	assert (name != NULL);

	/* Ignore call if no debug set. */
	if (get_gflag ())
	    {
		/* Write out a couple of extra descriptive stabs. */
		printf_sfile (".stabn\t192,0,0,.LBB%d-%s\n", be_seq, name);
		printf_sfile (".stabn\t224,0,0,.LBE%d-%s\n", be_seq, name);
		printf_sfile (".Lscope%d:\n", ++sc_seq);
		printf_sfile (".stabs\t\"\",36,0,0,");
		printf_sfile (".Lscope%d-%s\n", sc_seq, name);
	    }
}


/*
 * profile_function()
 *
 * Generate function profiling code.
 */
static void
profile_function (void)
{
	/* Ignore call if no profiling set. */
	if (get_pflag ())
	    {
		/* Write out code to link to the GNU profiling library. */
		printf_sfile (".data\n\t.align 4\n");
		printf_sfile (".LP%d:\n", ++lp_seq);
		printf_sfile ("\t.long 0\n");
		printf_sfile (".text\n");
		printf_sfile ("\tmovl $.LP%d,%%edx\n", lp_seq);
		printf_sfile ("\tcall mcount\n");
	    }
}


/*
 * generate_postamble()
 *
 * X86 debugging postamble generation routine.
 */
static void
generate_postamble (void)
{
	/* Don't write debug data if no debug set. */
	if (get_gflag ())
	    {
		/* Write out a few final comments. */
		printf_sfile ("\t.text\n");
		printf_sfile ("\t.stabs\t\"\",100,0,0,.Letext\n");
		printf_sfile (".Letext:\n");
	    }

	/* Write out common file postamble. */
	printf_sfile ("\t.ident\t \"Forthc: Forth 1.1, %s %s, Linux\"\n",
							__DATE__, __TIME__);
}


/*
 * generate_data()
 *
 * X86 string table generation routine.
 */
static void
generate_data (void)
{
	struct datarec	*datap;			/* Data record pointer */
	char		*cp;			/* Pointer to data content */

	/*
	 * Run through each pstore entry, finding definitions and generating
	 * code for each found.
	 */
	for (datap = p_datastart (); datap != NULL;
						datap = p_datanext (datap))
	    {
		/* Generate the header on the first time round. */
		if (datap == p_datastart ())
			printf_sfile (".section\t.rodata\n");

		/* Handle depending on the data type. */
		switch (datap->type)
		    {
			case P_STRING:
				printf_sfile (".LC%d:\n", datap->id);
				printf_sfile ("\t.string \"");
				for (cp = (char *) datap->data;
				     cp < (char *) datap->data + datap->len;
				     					cp++)
					printf_sfile ("%c", *cp);
				printf_sfile ("\"\n");
				break;

			case P_COUNTEDSTRING:
				printf_sfile (".LC%d:\n", datap->id);
				if (datap->len > 255)
					fatal ("counted string too long", "");
				printf_sfile ("\t.string \"\\%03o", datap->len);
				for (cp = (char *) datap->data;
				     cp < (char *) datap->data + datap->len;
				     					cp++)
					printf_sfile ("%c", *cp);
				printf_sfile ("\"\n");
				break;

			case P_DATA:
				printf_sfile (".LC%d:\n", datap->id);
				printf_sfile ("\t.string \"");
				for (cp = (char *) datap->data;
				     cp < (char *) datap->data + datap->len;
				     					cp++)
				    {
					if (*cp != '\0')
						printf_sfile ("%c", *cp);
					else
						printf_sfile
							("\"\n\t.string \"");
				    }
				printf_sfile ("\"\n");
				break;

			default:
				fatal ("unknown data table entry type", "");
		    }
	    }
}


/*
 * generate_variables()
 *
 * X86 variable tables generation routine.
 */
static void
generate_variables (void)
{
	struct symrec		*symp;		/* Symbol record pointer */

	/*
	 * Run through each pstore entry, finding definitions and generating
	 * code for each found.
	 */
	for (symp = p_symstart (); symp != NULL; symp = p_symnext (symp))
	    {
		/* Handle depending on the data type. */
		switch (symp->type)
		    {
			case P_VARIABLE:
				debug_variable (symp);
				printf_sfile ("\t.comm  %s,%d,4\n",
								symp->name,
					(symp->units == P_CELL) ?
					    symp->len * CELL_SIZE : symp->len);
				break;

			case P_CONSTANT:
				debug_variable (symp);
				printf_sfile ("\t.comm  %s,4,4\n",
								symp->name);
				break;

			case P_EXTERNALWORD:
			case P_WORD:
			case P_CODEWORD:
				break;

			default:
				fatal ("unknown symbol table entry type, %s",
								symp->name);
		    }
	    }
}


/*
 * map_register()
 *
 * Utility function to map a register number onto a string representing
 * an x86 CPU register.
 */
static const char *
map_register (int regnum)
{
	char	buf[32];			/* Integer conversion buffer */

	/* If the register number is invalid, just complain. */
	if (regnum < 0 || regnum >= NUM_REGISTERS)
	    {
		snprintf (buf, sizeof (buf), "%d", regnum);
		fatal ("map_register: invalid register, %s", buf);
	    }

	/* Return the corresponding register string. */
	return registers[regnum].x86reg;
}


/*
 * do_push()
 * do_pop()
 *
 * Generate code to push and pop to/from the selected register and stack.
 * 	push Rn,^x; pop Rn,^x
 */
static void
do_pushpop (const char *func, int regnum)
{
	const char	*x86reg;		/* Register name */
	assert (func != NULL);

	/* Identify the x86 register we need to use from regnum. */
	x86reg = map_register (regnum);

	/*
	 * If working with a register other than eax, exchange the registers'
	 * contents before, and after, the requested operation.  We do
	 * this because we know that eax is the register used by the the
	 * stack push and pop operations - if we are being asked to use
	 * another register with a stack, we need to shuffle data this way
	 * to preserve the current contents of eax.
	 */
	if (strcmp (x86reg, "%eax") != 0)
	    {
		printf_sfile ("\txchgl %%eax,%s\n", x86reg);
		printf_sfile ("\tcall %s\n", mangle (func));
		printf_sfile ("\txchgl %%eax,%s\n", x86reg);
	    }
	else
		/* Using register 0, so just call the requested operation. */
		printf_sfile ("\tcall %s\n", mangle (func));
}
static void
do_push (struct pcoderec *pcodep)
{
	char	*operands = pcodep->operands;	/* Operands */
	int	regnum;				/* Register number */
	char	stackid;			/* Stack ident */
	assert (operands != NULL);

	/* Scan in the operands. */
	if (sscanf (operands, "R%d,^%c", &regnum, &stackid) != 2)
		fatal ("push: scan error, %s", operands);

	/* Call the stack push function for the selected stack and register. */
	switch (stackid)
	    {
		case 'D':
			do_pushpop ("__fdpush", regnum);
			break;
		case 'R':
			do_pushpop ("__frpush", regnum);
			break;
		case 'F':
			do_pushpop ("__ffpush", regnum);
			break;
		default:
			fatal ("push: unknown stack id, %s", operands);
	    }
}
static void
do_pop (struct pcoderec *pcodep)
{
	char	*operands = pcodep->operands;	/* Operands */
	int	regnum;				/* Register number */
	char	stackid;			/* Stack ident */
	assert (operands != NULL);

	/* Scan in the operands. */
	if (sscanf (operands, "R%d,^%c", &regnum, &stackid) != 2)
		fatal ("pop: scan error, %s", operands);

	/* Call the stack push function for the selected stack and register. */
	switch (stackid)
	    {
		case 'D':
			do_pushpop ("__fdpop", regnum);
			break;
		case 'R':
			do_pushpop ("__frpop", regnum);
			break;
		case 'F':
			do_pushpop ("__ffpop", regnum);
			break;
		default:
			fatal ("pop: unknown stack id, %s", operands);
	    }
}


/*
 * do_load()
 *
 * Generate code to load a register with a specified value.
 * 	load Rn,$x; load Rn,#s; load Rn,@s; load Rn,'d; load Rn,Rn
 */
static void
do_load (struct pcoderec *pcodep)
{
	char	*operands = pcodep->operands;	/* Operands */
	int	regnum, num;			/* Register num/value */
	char	mode;				/* Addressing mode */
	const char	*x86reg;		/* Register name */
	struct symrec	*symp;			/* Symbol table entry */
	struct datarec	*datap;			/* Data table entry */
	assert (operands != NULL);

	/* Scan in the operands. */
	if (sscanf (operands, "R%d,%c%d", &regnum, &mode, &num) != 3)
		fatal ("load: scan error, %s", operands);
	x86reg = map_register (regnum);

	/* Handle the load depending on the mode of the second operand. */
	switch (mode)
	    {
		case '$':
			/* Load this value; optimize load of 0 to xorl x,x. */
			if (num == 0)
				printf_sfile ("\txorl %s,%s\n", x86reg, x86reg);
			else
				printf_sfile ("\tmovl $%d,%s\n", num, x86reg);
			break;

		case '#':
			/* Load the address of this symbol table entry. */
			symp = p_lookupsym (num);
			if (symp  == NULL)
				fatal ("load: symbol table entry missing, %s",
								operands);
			printf_sfile ("\tleal %s,%s\n", symp->name, x86reg);
			break;

		case '@':
			/* Load the data indexed by this symbol table entry. */
			symp = p_lookupsym (num);
			if (symp == NULL)
				fatal ("load: symbol table entry missing, %s",
								operands);
			printf_sfile ("\tmovl (%s),%s\n", symp->name, x86reg);
			break;

		case '\'':
			/* Load the address of this data table entry. */
			datap = p_lookupdata (num);
			if (datap == NULL)
				fatal ("load: data table entry missing, %s",
								operands);
			printf_sfile ("\tleal .LC%d,%s\n", datap->id, x86reg);
			break;

		case 'R':
			/* Load the contents of the target register. */
			printf_sfile ("\tmovl %s,%s\n", map_register (num),
									x86reg);
			break;

		default:
			fatal ("load: unknown operand mode, %s", operands);
	    }
}


/*
 * do_store()
 *
 * Store the value of a register in a given location.
 * 	store Rn,@s
 */
static void
do_store (struct pcoderec *pcodep)
{
	char	*operands = pcodep->operands;	/* Operands */
	int	regnum, num;			/* Register num/value */
	char	mode;				/* Addressing mode */
	const char	*x86reg;		/* Register name */
	struct symrec	*symp;			/* Symbol table entry */
	assert (operands != NULL);

	/* Scan in the operands. */
	if (sscanf (operands, "R%d,%c%d", &regnum, &mode, &num) != 3)
		fatal ("store: scan error, %s", operands);
	x86reg = map_register (regnum);

	/* Handle according to the operand mode. */
	switch (mode)
	    {
		case '@':
			/* Store in data indexed by this symbol table entry. */
			symp = p_lookupsym (num);
			if (symp == NULL)
				fatal ("store: symbol table entry missing, %s",
								operands);
			printf_sfile ("\tmovl %s,(%s)\n", x86reg, symp->name);
			break;

		default:
			fatal ("store: unknown operand mode, %s", operands);
	    }
}


/*
 * do_jump()
 *
 * Jump, conditional or unconditional, to the given label number.  Jump
 * comparisons are restricted to registers only.  Two registers compares
 * the values against each other.  A single register just tests the value
 * of that register.
 * 	jump $l; jump_eq Rn,[Rn,]$l; jump_ne Rn,[Rn,]$l; jump_gt Rn,[Rn,]$l;
 * 	jump_ge Rn,[Rn,]$l...
 */
static void
do_jump (struct pcoderec *pcodep)
{
	char	*opcode = pcodep->opcode;	/* Specific opcode */
	char	*operands = pcodep->operands;	/* Operands */
	int	regnum;				/* Register number */
	int	regnum1, regnum2;		/* Register numbers */
	int	num;				/* Label number */
	static const struct jumprec {		/* Maps jumps to x86 opcodes */
		const char	*jump;
		const char	*x86jump;
	} jumptab[] = {
	{ "jump",    "jmp" },
	{ "jump_nz", "jne" }, { "jump_ne", "jne" }, { "jump_z",  "je"  },
	{ "jump_eq", "je"  }, { "jump_gt", "jg"  }, { "jump_lt", "jl"  },
	{ "jump_ge", "jge" }, { "jump_le", "jle" }, { NULL,       NULL } };
	const struct jumprec	*jrecp;		/* Pointer to jumptab entry */
	assert (opcode != NULL && operands != NULL);

	/*
	 * If we have two registers, compare one against the other; if one,
	 * test it's value; and if none, then this is an unconditional
	 * jump (check the opcode for this condition before proceeding).
	 */
	if (sscanf (operands, "R%d,R%d,$%d", &regnum1, &regnum2, &num) == 3)
		printf_sfile ("\tcmpl %s,%s\n", map_register (regnum1),
						map_register (regnum2));
	else
	    {
		if (sscanf (operands, "R%d,$%d", &regnum, &num) == 2)
			printf_sfile ("\ttestl %s,%s\n", map_register (regnum),
							 map_register (regnum));
		else
		    {
			if (strcmp (opcode, "jump") == 0)
			    {
				if (sscanf (operands, "$%d", &num) != 1)
					fatal ("jump: jump scan error, %s",
								operands);
			    }
			else
				fatal ("jump: scan error, %s", operands);
		    }
	    }

	/*
	 * Generate the jump according to the specific opcode.  For condit-
	 * ional jumps, we search the table looking for a match against
	 * the specific jump opcode we picked up from the pcode pointer.
	 */
	for (jrecp = jumptab; jrecp->jump != NULL; jrecp++)
	    {
		if (strcmp (opcode, jrecp->jump) == 0)
		    {
			/* Found the entry; generate the jump and break. */
			printf_sfile ("\t%s .L%d\n", jrecp->x86jump, num);
			break;
		    }
	    }

	/* If no matching jump was found, this is an error. */
	if (jrecp == NULL)
		fatal ("jump: invalid jump opcode, %s", opcode);
}


/*
 * do_add()
 * do_subtract()
 *
 * Add the contents of a register, or an absolute value, to a register.
 * 	add Rn,Rn; add Rn,$x
 */
static void
do_add (struct pcoderec *pcodep)
{
	char	*operands = pcodep->operands;	/* Operands */
	int	regnum, num;			/* Register num/value */
	char	mode;				/* Addressing mode */
	const char	*x86reg;		/* Register name */
	assert (operands != NULL);

	/* Scan in the operands. */
	if (sscanf (operands, "R%d,%c%d", &regnum, &mode, &num) != 3)
		fatal ("add: scan error, %s", operands);
	x86reg = map_register (regnum);

	/* Handle according to the second operand mode. */
	switch (mode)
	    {
		case '$':
			/* Add the immediate value to the given register. */
			switch (num)
			    {
				case  1:
					printf_sfile ("\tincl %s\n", x86reg);
					break;
				case -1:
					printf_sfile ("\tdecl %s\n", x86reg);
					break;
				case  0:
					break;
				default:
					printf_sfile ("\taddl $%d,%s\n",
								num, x86reg);
					break;
			    }
			break;

		case 'R':
			/* Add the second register to the first. */
			printf_sfile ("\taddl %s,%s\n", map_register (num),
								     x86reg);
			break;

		default:
			fatal ("add: unknown operand mode, %s", operands);
	    }
}
static void
do_subtract (struct pcoderec *pcodep)
{
	char	*operands = pcodep->operands;	/* Operands */
	int	regnum, num;			/* Register num/value */
	char	mode;				/* Addressing mode */
	const char	*x86reg;		/* Register name */
	assert (operands != NULL);

	/* Scan in the operands. */
	if (sscanf (operands, "R%d,%c%d", &regnum, &mode, &num) != 3)
		fatal ("subtract: scan error, %s", operands);
	x86reg = map_register (regnum);

	/* Handle according to the second operand mode. */
	switch (mode)
	    {
		case '$':
			/* Subtract the immediate value from given register. */
			switch (num)
			    {
				case  1:
					printf_sfile ("\tdecl %s\n", x86reg);
					break;
				case -1:
					printf_sfile ("\tincl %s\n", x86reg);
					break;
				case  0:
					break;
				default:
					printf_sfile ("\taddl $%d,%s\n",
								num, x86reg);
					break;
			    }
			break;

		case 'R':
			/* Subtract the second register from the first. */
			printf_sfile ("\tsubl %s,%s\n", map_register (num),
								     x86reg);
			break;

		default:
			fatal ("subtract: unknown operand mode, %s", operands);
	    }
}


/*
 * do_call
 *
 * Call the symbol identified in the operands.
 * 	call #s
 */
static void
do_call (struct pcoderec *pcodep)
{
	char	*operands = pcodep->operands;	/* Operands */
	int	num;				/* Symbol entry id */
	struct symrec	*symp;			/* Symbol table entry */
	assert (pcodep != NULL);

	/* Scan in the operands. */
	if (sscanf (operands, "#%d", &num) != 1)
		fatal ("call: scan error, %s", operands);

	/* Check that the symbol is a valid callable function. */
	symp = p_lookupsym (num);
	if (symp == NULL)
		fatal ("call: symbol table entry missing, %s", operands);

	/* Call the function by name. */
	printf_sfile ("\tcall %s\n", symp->name);
}


/*
 * do_label()
 *
 * Generate labels.
 * 	label $l
 */
static void
do_label (struct pcoderec *pcodep)
{
	char	*operands = pcodep->operands;	/* Operands */
	int	num;				/* Label number */
	assert (operands != NULL);

	/* Scan in the operand. */
	if (sscanf (operands, "$%d", &num) != 1)
		fatal ("label: scan error, %s", operands);

	/* Generate the label specified by this call. */
	printf_sfile (".L%d:\n", num);
}


/*
 * do_define()
 * do_enddef()
 *
 * Generate definition starts, and definition ends.
 * 	define #s; enddef
 */
static void
do_define (struct pcoderec *pcodep)
{
	char	*operands = pcodep->operands;	/* Operands */
	int	num;				/* Symbol entry id */
	int	regnum;				/* Register number for loop */
	struct symrec	*symp;			/* Symbol table entry */
	assert (pcodep != NULL);

	/* Scan in the operand. */
	if (sscanf (operands, "#%d", &num) != 1)
		fatal ("define: scan error, %s", operands);

	/* Check the symbol is a valid function symbol. */
	symp = p_lookupsym (num);
	if (symp == NULL)
		fatal ("define: symbol table entry missing, %s", operands);

	/* Housekeeping on definition entry. */
	currdef = symp;
	defcount++;

	/* Write the definition. */
	printf_sfile (".text\n");
	printf_sfile ("\t.align 4\n");
	debug_function (currdef->name);
	printf_sfile (".globl %s\n", currdef->name);
	printf_sfile ("\t.type\t%s,@function\n", currdef->name);
	printf_sfile ("%s:\n", currdef->name);

	/*
	 * Set up C-like stack frame pointer, so gdb can understand it in
	 * the event of core dumps.
	 */
	debug_statement (pcodep->linenum);
	printf_sfile ("\tpushl %%ebp\n");
	printf_sfile ("\tmovl %%esp,%%ebp\n");

	/* Preserve used registers for words - codewords do their own. */
	if (currdef->type == P_WORD)
	    {
		/* Scan forward on registers, pushing all marked preserved. */
		for (regnum = 0; regnum < NUM_REGISTERS; regnum++)
		    {
			if (registers[regnum].preserved)
				printf_sfile ("\tpushl %s\n",
						registers[regnum].x86reg);
		    }
	    }

	/* Emit profiling and debugging code. */
	profile_function ();
	debug_statement (pcodep->linenum);
	debug_function_begin (currdef->name);

	/* If this is a code definition, hint to the assembler. */
	if (currdef->type == P_CODEWORD)
		printf_sfile ("#APP\n");
}
static void
do_enddef (struct pcoderec *pcodep)
{
	int	regnum;				/* Register number for loop */
	assert (pcodep != NULL);

	/* If this was a code definition, undo the assembler hint. */
	if (currdef->type == P_CODEWORD)
		printf_sfile ("#NO_APP\n");

	/* Unravel stack frame. */
	debug_function_end (currdef->name);
	debug_statement (pcodep->linenum);

	/* Restore used registers for words. */
	if (currdef->type == P_WORD)
	    {
		/* Scan backward on registers, popping all marked preserved. */
		for (regnum = NUM_REGISTERS - 1; regnum >= 0; regnum--)
		    {
			if (registers[regnum].preserved)
				printf_sfile ("\tpopl %s\n",
						registers[regnum].x86reg);
		    }
	    }

	/* Clean up the stack and return. */
	printf_sfile ("\tleave\n");
	printf_sfile ("\tret\n");

	/* Complete function postamble. */
	printf_sfile (".Lfe%d:\n", defcount);
	printf_sfile ("\t.size\t %s,.Lfe%d-%s\n",
				currdef->name, defcount, currdef->name);
	debug_function_cleanup (currdef->name);

	/* Housekeeping on definition exit. */
	currdef = NULL;
}


/*
 * do_asm()
 *
 * Pass through assembly language to the sfile.
 * 	asm <string>
 */
static void
do_asm (struct pcoderec *pcodep)
{
	char		*operands= pcodep->operands;	/* Operands */
	assert (operands != NULL);

	/* Write the assembler through to the sfile. */
	printf_sfile ("%s\n", operands);
}


/*
 * do_main_start()
 * do_main_end()
 *
 * Generate a main function; called if needed.
 */
static void
do_main_start (int linenum)
{
	int	regnum;				/* Register number for loop */

	/* Write general function preamble. */
	printf_sfile (".text\n");
	printf_sfile ("\t.align 4\n");
	debug_function ("main");
	printf_sfile (".globl main\n");
	printf_sfile ("\t.type\tmain,@function\n");
	printf_sfile ("main:\n");
	debug_statement (linenum);
	printf_sfile ("\tpushl %%ebp\n");
	printf_sfile ("\tmovl %%esp,%%ebp\n");

	/* Scan forward on registers, pushing all marked preserved. */
	for (regnum = 0; regnum < NUM_REGISTERS; regnum++)
	    {
		if (registers[regnum].preserved)
			printf_sfile ("\tpushl %s\n", registers[regnum].x86reg);
	    }

	/* Copy argc, argv, and envp from the stack to global variables. */
	printf_sfile ("\tmovl 0x08(%%ebp),%%eax\n\tmovl %%eax,_argc\n");
	printf_sfile ("\tmovl 0x0C(%%ebp),%%eax\n\tmovl %%eax,_argv\n");
	printf_sfile ("\tmovl 0x10(%%ebp),%%eax\n\tmovl %%eax,_envp\n");

	/* Profile, if needed, and increment definitions count. */
	profile_function ();
	debug_statement (linenum);
	debug_function_begin ("main");
	defcount++;
}
static void
do_main_end (int linenum)
{
	int	regnum;				/* Register number for loop */

	/* Write general postamble. */
	debug_function_end ("main");
	debug_statement (linenum);

	/* Scan backward on registers, popping all marked preserved. */
	for (regnum = NUM_REGISTERS - 1; regnum >= 0; regnum--)
	    {
		if (registers[regnum].preserved)
			printf_sfile ("\tpopl %s\n", registers[regnum].x86reg);
	    }

	/* Clean up the stack and return. */
	printf_sfile ("\tleave\n");
	printf_sfile ("\tret\n");

	/* Complete main function postamble. */
	printf_sfile (".Lfe%d:\n", defcount);
	printf_sfile ("\t.size\t main,.Lfe%d-main\n", defcount);
	debug_function_cleanup ("main");
}


/* Table of functions handling pcode opcode. */
#define DEFINE		"define"
#define ENDDEF		"enddef"
static const struct oprec {
	const char	*opcode;
	void		(*const func)( struct pcoderec *pcodep );
	const int	executable;
} optable[] = {
	{ "call",    do_call,   TRUE },  { "no_op",    NULL,        TRUE },
	{ "load",    do_load,   TRUE },  { "store",    do_store,    TRUE },
	{ "push",    do_push,   TRUE },  { "pop",      do_pop,      TRUE },
	{ "add",     do_add,    TRUE },  { "subtract", do_subtract, TRUE },
	{ "jump",    do_jump,   TRUE },
	{ "jump_z",  do_jump,   TRUE },  { "jump_nz",  do_jump,     TRUE },
	{ "jump_eq", do_jump,   TRUE },  { "jump_ne",  do_jump,     TRUE },
	{ "jump_gt", do_jump,   TRUE },  { "jump_lt",  do_jump,     TRUE },
	{ "jump_ge", do_jump,   TRUE },  { "jump_le",  do_jump,     TRUE },
	{ "label",   do_label,  FALSE },
	{ DEFINE,    do_define, FALSE }, { ENDDEF,     do_enddef,   FALSE },
	{ "asm",     do_asm,    FALSE },
	{ NULL,      NULL,      FALSE }
};


/*
 * despatch()
 *
 * Lookup an opcode, and despatch the handler.
 */
static void
despatch (struct pcoderec *pcodep, int debug)
{
	const struct oprec	*opp;		/* Pointer through opcodes */
	assert (pcodep != NULL);

	/* Look up the pcode opcode in the table. */
	for (opp = optable; opp->opcode != NULL; opp++)
	    {
		if (strcmp (opp->opcode, pcodep->opcode) == 0)
		    {
			/* First, emit debug if there is associated source. */
			if (debug && pcodep->source_token && opp->executable)
				debug_statement (pcodep->linenum);

			/* Now call the handler function if not NULL. */
			if (opp->func != NULL)
				(opp->func) (pcodep);
			break;
		    }
	    }

	/* Fatal error if the opcode wasn't found. */
	if (opp->opcode == NULL)
		fatal ("despatch: unknown opcode, %s", pcodep->opcode);
}


/*
 * generate_code()
 *
 * Generate x86 assembler for code in the pstore.
 */
static void
generate_code (void)
{
	struct pcoderec	*pcodep;		/* Pointer through pcode list */
	struct pcoderec	*main_pcodep = NULL;	/* Pointer to any main start */
	int		in_definition = FALSE;	/* If in definition flag */
	int		currline;		/* Line number change test */

	/*
	 * Run through each pcode entry, finding definitions and generating
	 * code for each found.  At the same time, if we find pcode that is
	 * not in a definition, we set the main pcodep to the first of these.
	 */
	currline = -1; in_definition = FALSE;
	for (pcodep = p_codestart (); pcodep != NULL;
						pcodep = p_codenext (pcodep))
	    {
		/* See if this is the start of a definition. */
		if (strcmp (pcodep->opcode, DEFINE) == 0)
			in_definition = TRUE;

		/*
		 * Invoke the correct code generation function if inside
		 * a definition section, with debug only on linenum change.
		 */
		if (in_definition)
		    {
			despatch (pcodep, currline != pcodep->linenum);
			currline = pcodep->linenum;
		    }
		else
		    {
			/*
			 * If this opcode is not define or enddef, it is code.
			 * If it is the first code we've seen, note as 'main'.
			 */
			if (strcmp (pcodep->opcode, DEFINE) != 0
					&& strcmp (pcodep->opcode, ENDDEF) != 0)
				if (main_pcodep == NULL)
					main_pcodep = pcodep;
		    }

		/* See if this ends a definition. */
		if (strcmp (pcodep->opcode, ENDDEF) == 0)
			in_definition = FALSE;
	    }

	/*
	 * Now make a second pass, starting at the main code pcode entry, if
	 * found, treating anything outside a definition as the function
	 * 'main'.
	 */
	if (main_pcodep != NULL)
	    {
		/* Generate the 'main' function introduction. */
		do_main_start (main_pcodep->linenum);

		/* Loop through pcode, starting at the 'main' first entry. */
		currline = -1; in_definition = FALSE;
		for (pcodep = main_pcodep; pcodep != NULL;
						pcodep = p_codenext (pcodep))
		    {
			/* See if this is the start or end of a definition. */
			if (strcmp (pcodep->opcode, DEFINE) == 0)
			    {
				in_definition = TRUE;
				continue;
			    }
			else
				if (strcmp (pcodep->opcode, ENDDEF) == 0)
				    {
					in_definition = FALSE;
					continue;
				    }

			/* Proceed only if not in a definition. */
			if (!in_definition)
			    {
				/* Invoke the right code generation function. */
				despatch (pcodep, currline != pcodep->linenum);
				currline = pcodep->linenum;
			    }
		    }

		/*
		 * Currline contains the line number of the last line for
		 * which we generated any code.
		 */
		do_main_end (currline);
	}
}


/*
 * generate_assembler()
 *
 * Main entry point for this module.  Generate an x86 assembler file for
 * the symbols, data, and code in the pstore.
 */
void
generate_assembler (const char *source_path)
{
	assert (source_path != NULL);

	/* Clear down count and current definition from any prior call. */
	defcount	= 0;
	currdef		= NULL;

	/* Clear sequential counts of debugging and begin-end labels. */
	lp_seq		= 0;
	lm_seq		= 0;
	be_seq		= 0;
	sc_seq		= 0;

	/*
	 * Simply run each code generation phase in sequence, starting
	 * with the preamble, adding data and variables, writing code,
	 * then rounding off with the postamble.
	 */
	generate_preamble (source_path);
	generate_data ();
	generate_variables ();
	generate_code ();
	generate_postamble ();
}
