/*
 * VNPForth - Compiled native Forth for IA32 Linux systems
 * Copyright (C) 2001  Simon Baldwin (simonb@caldera.com)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

%{
#include <assert.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <malloc.h>
#include <errno.h>

#include "ident.h"
#include "ifile.h"
#include "pstore.h"

#include "forth_y.h"

/* Definitions. */
#define YYSTYPE			char*
#define YYERROR_VERBOSE		1
#define	DESCENT_STACK_SIZE	1024
#define	MAX_ALLOWED_ERRORS	20

/* True/False macros. */
#define FALSE			(0)
#define TRUE			(!FALSE)

/*
 * Macro to cause yyparse to abort if the maximum number of allowable errors
 * has been exceeded.  This is an icky thing to have to do, but the closed
 * nature of yyparse seems to leave little option - even a setjmp/longjmp is
 * not an improvement.  And it doesn't even trigger until we hit a valid
 * token, so we may well have a lot more errors than the maximum allowed by
 * the time that happens.  All in all, rather unpleasant.
 */
#define	MNE	do {							\
			if (yy_errcount >= MAX_ALLOWED_ERRORS)		\
			    {						\
				yyerror ("too many errors, giving up");	\
				YYABORT;				\
			    }						\
		} while (0)


/* Count of errors and warnings found in the file. */
int		yy_errcount	= 0;
int		yy_warncount	= 0;

/* Flag to help us tell if we are inside yyparse, or not. */
static int	in_yyparse	= FALSE;

/* Line number from lex, and lex infile stream reset function. */
extern int	yy_linenum;
extern void	yyrestart (FILE*);

/* External declaration of yylex lexical tokenizer. */
extern int	yylex (void);

/* Forward declarations of error and parse terminal handlers. */
void		yyerror (const char *message);
static void	f_definedvariable (char *s);
static void	f_definedarray (char *s, char *ssize, char *type);
static void	f_definedconstant (char *s);
static void	f_definedword (char *s);
static void	f_semicolon (char *s);
static void	f_exit (char *s);
static void	f_pstring (char *s);
static void	f_cstring (char *s);
static void	f_word (char *s);
static void	f_object_address (char *s);
static void	f_char_literal (char *s);
static void	f_integer_literal (char *s);
static void	f_float_literal (char *s);
static void	f_string_literal (char *s);
static void	f_if (char *s);
static void	f_then (char *s);
static void	f_else (char *s);
static void	f_do (char *s);
static void	f_query_do (char *s);
static void	f_loop (char *s);
static void	f_plus_loop (char *s);
static void	f_leave (char *s);
static void	f_begin (char *s);
static void	f_until (char *s);
static void	f_again (char *s);
static void	f_while (char *s);
static void	f_repeat (char *s);
static void	f_case (char *s);
static void	f_of (char *s);
static void	f_endof (char *s);
static void	f_endcase (char *s);
static void	f_recurse (char *s);
static void	f_codeword (char *s);
static void	f_code (char *s);
static void	f_end_code (char *s);
%}

%token VARIABLE CONSTANT
%token CREATE ALLOT
%token EXIT
%token IF THEN ELSE
%token DO QUERY_DO "?DO" LOOP PLUS_LOOP "+LOOP" LEAVE
%token _BEGIN "BEGIN" UNTIL WHILE REPEAT AGAIN
%token CASE OF ENDOF ENDCASE
%token RECURSE
%token PSTRING "print string"
%token CSTRING "counted string"
%token INTEGER_LITERAL "integer"
%token STRING_LITERAL "literal string"
%token FLOAT_LITERAL "float"
%token CHAR_LITERAL "char literal"
%token OBJECT_ADDRESS "address"
%token WORD "word"
%token CODE ASM "assembler string" END_CODE "END-CODE"

%%
program:		elements
			;

elements:		/* Empty. */
			| elements element
			;

element:		declaration
			| definition
			| statement
			| error	WORD		{ MNE; yyerrok; }
			;

declaration:		VARIABLE WORD		{ MNE; f_definedvariable ($2); }
			| CONSTANT WORD		{ MNE; f_definedconstant ($2); }
			| CREATE WORD INTEGER_LITERAL WORD ALLOT { MNE;
						f_definedarray ($2, $3, $4); }
			| CREATE WORD INTEGER_LITERAL ALLOT { MNE;
						f_definedarray ($2, $3, NULL);}
			;

definition:		':' definedword statements _semicolon
			| CODE codeword codes _end_code
			;
definedword:		WORD			{ MNE; f_definedword ($1); }
			;
_semicolon:		';'			{ MNE; f_semicolon ($1); }
			;
codeword:		WORD			{ MNE; f_codeword ($1); }
			;
codes:			code
			| codes code
			;
code:			ASM			{ MNE; f_code ($1); }
			;
_end_code:		END_CODE		{ MNE; f_end_code ($1); }
			;

statement:		if_statement
			| do_loop
			| do_leave
			| begin_loop
			| begin_while
			| case_statement
			| recurse
			| EXIT			{ MNE; f_exit ($1); }
			| PSTRING		{ MNE; f_pstring ($1); }
			| CSTRING		{ MNE; f_cstring ($1); }
			| INTEGER_LITERAL	{ MNE; f_integer_literal ($1); }
			| FLOAT_LITERAL		{ MNE; f_float_literal ($1); }
			| STRING_LITERAL	{ MNE; f_string_literal ($1); }
			| CHAR_LITERAL		{ MNE; f_char_literal ($1); }
			| OBJECT_ADDRESS	{ MNE; f_object_address ($1); }
			| ALLOT			{ MNE; f_word ($1); }
			| WORD			{ MNE; f_word ($1); }
			;

if_statement:		_if statements _then
			| _if statements _else statements _then
			;
_if:			IF			{ MNE; f_if ($1); }
			;
_then:			THEN			{ MNE; f_then ($1); }
			;
_else:			ELSE			{ MNE; f_else ($1); }
			;

do_loop:		_do statements _loop
			| _do statements _plus_loop
			;
_do:			DO			{ MNE; f_do ($1); }
			| QUERY_DO 		{ MNE; f_query_do ($1); }
			;
_loop:			LOOP			{ MNE; f_loop ($1); }
			;
_plus_loop:		PLUS_LOOP		{ MNE; f_plus_loop ($1); }
			;
do_leave:		LEAVE			{ MNE; f_leave ($1); }
			;

begin_loop:		_begin statements _until
			| _begin statements _again
			| _begin statements _repeat
			;
_begin:			_BEGIN			{ MNE; f_begin ($1); }
			;
_until:			UNTIL			{ MNE; f_until ($1); }
			;
_again:			AGAIN			{ MNE; f_again ($1); }
			;
_repeat:		REPEAT			{ MNE; f_repeat ($1); }
			;
begin_while:		WHILE			{ MNE; f_while ($1); }
			;

case_statement:		_case casebody statements _endcase
			;
casebody:		caseclause
			| casebody caseclause
			;
caseclause:		statements _of statements _endof
			;
_case:			CASE			{ MNE; f_case ($1); }
			;
_endcase:		ENDCASE			{ MNE; f_endcase ($1); }
			;
_of:			OF			{ MNE; f_of ($1); }
			;
_endof:			ENDOF			{ MNE; f_endof ($1); }
			;

recurse:		RECURSE			{ MNE; f_recurse ($1); }
			;

statements:		elements
			;
%%


/*
 * do_message()
 * yyerror[a](),yywarning[a]()
 *
 * Parse error functions, also called from lexical analyzer, one
 * offering a bit more error detail, and parse warning functions.
 */
static void
do_message (const char *format, const char *message)
{
	assert (format != NULL && message != NULL);

	/*
	 * If in_yyparse is false then we are not in the process of parsing
	 * a file, so there is no valid "near '...'" to print.  Otherwise,
	 * show the latest token found as a reference point in the source.
	 */
	if (in_yyparse)
		fprintf (stderr, "%s:%d: near '%s', ",
					get_ifname (), yy_linenum, yylval);
	else
		/* We are not "near" anything. */
		fprintf (stderr, "%s: ", get_ifname ());

	/* Print the error details. */
	fprintf (stderr, format, message);
	fprintf (stderr, "\n");
}
void
yyerrora (const char *format, const char *message)
{
	do_message (format, message);
	yy_errcount++;
}
void
yyerror (const char *message)
{
	yyerrora ("%s", message);
}
void
yywarninga (const char *format, const char *message)
{
	do_message (format, message);
	yy_warncount++;
}
void
yywarning (const char *message)
{
	yywarninga ("%s", message);
}


/*
 * Parse descent (label) stack, index, and next label counter; label 0 is
 * reserved for 'main' code.
 */
static int	label_stack[DESCENT_STACK_SIZE];
static int	label_index	= 0;
static int	label_seq	= 1;

/*
 * labeltos(),labelpush(),labelpop()
 *
 * Definitions for pushing a label onto the stack, and popping one off.
 */
#ifdef FOR_FUTURE_EXPANSION
static int
labeltos (void)
{
	if (label_index <= 0)
		yyerror ("internal error: parsing stack underrun");
	return label_stack[label_index - 1];
}
#endif
static int
labelpush (void)
{
	if (label_index >= DESCENT_STACK_SIZE - 1)
		yyerror ("internal error: statement nesting too deep");
	label_stack[label_index++] = label_seq++;
	return label_stack[label_index - 1];
}
static int
labelpop (void)
{
	if (label_index <= 0)
		yyerror ("internal error: parsing stack underrun");
	return label_stack[--label_index];
}

/*
 * Note of the label number to emit at the end of the current function
 * definition - zero is special for the 'main' function.
 */
static int	end_label	= 0;
static int	need_main_end	= FALSE;

/* Flag to indicate this module contains 'main'. */
static int	has_main	= FALSE;

/* Pointer to the symbol table entry of any currently entered definition. */
static struct symrec	*currdef	= NULL;


/*
 * Break/leave stack, and current stack location/index pointer.
 */
static int	break_stack[DESCENT_STACK_SIZE];
static int	break_index	= 0;

/*
 * breaktos(),breakpush(),breakpop()
 *
 * Loop/case labels nesting stack, for leave command, and endcase break label.
 */
static int
breaktos (void)
{
	if (break_index <= 0)
		yyerror ("internal error: loop stack underrun");
	return break_stack[break_index - 1];
}
static void
breakpush (int label_number)
{
	if (break_index >= DESCENT_STACK_SIZE - 1)
		yyerror ("internal error: loop nesting too deep");
	break_stack[break_index++] = label_number;
}
static int
breakpop (void)
{
	if (break_index <= 0)
		yyerror ("internal error: loop stack underrun");
	return break_stack[--break_index];
}


/*
 * ltoa(),adjstr(),skipspace(),skipnonspace(),truncspace()
 *
 * General utility functions for parse handlers.
 */
#define	GENERAL_BUF_LEN		32
static char *
ltoa (int i)
{
	static char	buf[GENERAL_BUF_LEN];

	/* Return a label target string from an integer target. */
	snprintf (buf, sizeof (buf), "$%d", i);
	return buf;
}
static char *
adjstr (char *s)
{
	assert (s != NULL);

	/* Adjust a string by removing the initial x"space, and trailing ". */
	*(s + strlen(s) - 1) = '\0';
	return s + strlen ("x\" ");
}
static char *
skipspace (char *s)
{
	char		*sp = s;
	assert (s != NULL);

	/* Skip leading whitespace. */
	while (isspace (*sp))
		sp++;
	return sp;
}
static char *
skipnonspace (char *s)
{
	char		*sp = s;
	assert (s != NULL);

	/* Skip leading non-whitespace. */
	while (!isspace (*sp))
		sp++;
	return sp;
}
static void
truncspace (char *s)
{
	char		*sp = skipnonspace (s);
	assert (s != NULL);

	/* Truncate to the first space char found. */
	*sp = '\0';
}


/*
 * wp_append()
 *
 * Function to detect whether this module contains 'main' - the condition
 * for containing 'main' is detecting a statement outside of any definition;
 * this wrapper to p_append detects this condition.  It also adds the
 * current lexical line number, the value of yy_linenum, to save us adding
 * it at each call.
 */
static void
wp_append (char *opcode, char *operand, char *source_token)
{
	assert (opcode != NULL);
	if (currdef == NULL)
		has_main = TRUE;
	p_append (opcode, operand, source_token, yy_linenum);
}


/*
 * f_definedcommon(),f_definedvariable(),f_definedarray()
 * f_definedconstant()
 *
 * Parse terminals handlers for variable definitions.
 */
static void
f_definedcommon (char *s, int size, enum elunits units)
{
	struct symrec	*symp = p_findsym (mangle (s));

	/* Add variable to symbol table if not yet defined. */
	if (size <= 0)
		yyerrora ("item '%s' cannot have size <= 0", s);
	else
	    {
		if (symp != NULL)
		    {
			if (symp->type != P_EXTERNALWORD)
				yyerrora ("multiple symbol definition '%s'", s);
			else
				yyerrora
				 ("symbol already implicitly external '%s'", s);
		    }
		p_addsym (P_VARIABLE, mangle (s), size, units);
	    }
}
static void
f_definedvariable (char *s)
{
	assert (s != NULL);

	/* Define a new single cell. */
	f_definedcommon (s, 1, P_CELL);
}
static void
f_definedarray (char *s, char *ssize, char *type)
{
	assert (s != NULL && ssize != NULL);

	/* Define an array of cells/chars; NULL type implies chars. */
	truncspace (s); truncspace (ssize);
	if (type == NULL)
		f_definedcommon (s, strtol (ssize, NULL, 0), P_CHAR);
	else
	    {
		truncspace (type);
		if (strcmp (type, "CHARS") == 0)
			f_definedcommon (s, strtol (ssize, NULL, 0), P_CHAR);
		else
		    {
			if (strcmp (type, "CELLS") == 0)
				f_definedcommon (s, strtol (ssize, NULL, 0),
									P_CELL);
			else
				yyerror
				  ("parse error, expecting 'CHARS' or 'CELLS'");
		    }
	    }
}
#ifdef FOR_FUTURE_EXPANSION
static void
f_definedchararray (char *s, char *ssize)
{
	assert (s != NULL && ssize != NULL);

	/* Define an char array. */
	truncspace (s); truncspace (ssize);
	f_definedcommon (s, strtol (ssize, NULL, 0), P_CHAR);
}
#endif
static void
f_definedconstant (char *s)
{
	char		buf[GENERAL_BUF_LEN];
	struct symrec	*symp = p_findsym (mangle (s));

	/* Add constant to symbol table if not yet defined. */
	if (symp != NULL ) 
	    {
		if (symp->type != P_EXTERNALWORD)
			yyerrora ("multiple symbol definition '%s'", s);
		else
			yyerrora ("symbol already implicitly external '%s'", s);
	    }
	symp = p_addsym (P_CONSTANT, mangle (s), 1, P_CELL);
	wp_append ("pop", "R0,^D", s);
	snprintf (buf, sizeof (buf), "R0,@%d", symp->id);
	wp_append ("store", buf, NULL);
}


/*
 * f_definedword(),f_semicolon(),f_exit(),f_recurse
 *
 * Parse terminal handlers for word definition, word exit, and recurse.
 */
static void
f_definedword (char *s)
{
	char		buf[GENERAL_BUF_LEN];
	struct symrec	*symp = p_findsym (mangle (s));

	/* Add word to symbol table if not yet defined, or undefined. */
	if (symp != NULL)
	    {
		/* If undefined, change the symbol to be a defined word. */
		if (symp->type != P_EXTERNALWORD)
			yyerrora ("multiple symbol definition '%s'", s);
		else
			p_chgsym (symp, P_WORD, 1, P_CELL);
	    }
	else
		/* Add the new symbol for this word. */
		symp = p_addsym (P_WORD, mangle (s), 1, P_CELL);

	/* Set as current definition, and add pcode for a define. */
	currdef = symp;
	end_label = labelpush ();
	snprintf (buf, sizeof (buf), "#%d", symp->id);
	wp_append ("define", buf, s);
}
static void
f_semicolon (char *s)
{
	assert (s != NULL);

	/* End a word definition. */
	wp_append ("label", ltoa (labelpop ()), s);
	wp_append ("enddef", NULL, NULL);
	end_label = 0;
	currdef = NULL;
}
static void
f_exit (char *s)
{
	assert (s != NULL);

	/* Exit prematurely from this definition/function. */
	wp_append ("jump", ltoa (end_label), s);
	if (end_label == 0)
		need_main_end = TRUE;
}
static void
f_recurse (char *s)
{
	char		buf[GENERAL_BUF_LEN];
	assert (s != NULL);

	/* Call the routine we are already defining - for compatibility. */
	if (currdef == NULL)
		yyerrora ("'%s' invalid outside a definition", s);
	else
	    {
		snprintf (buf, sizeof (buf), "#%d", currdef->id);
		wp_append ("call", buf, s);
	    }
}


/*
 * f_pstring(),f_cstring(),f_string_literal
 *
 * Parse terminal handlers for printed strings, counted strings, and string
 * literals.
 */
static void
f_pstring (char *s)
{
	char		buf[GENERAL_BUF_LEN];
	char		*adjs = adjstr (s);
	struct datarec	*datap;
	struct symrec	*symp = p_findsym (mangle ("TYPE"));

	/*
	 * Add a string to data table, and emit code to print it.  First, add
	 * a symbol table definition for "TYPE" if necessary.
	 */
	if (symp != NULL)
	    {
		if (symp->type != P_EXTERNALWORD
				&& symp->type != P_WORD
				&& symp->type != P_CODEWORD)
			yyerrora ("'%s' is defined as a variable", "TYPE");
	    }
	else
		symp = p_addsym (P_EXTERNALWORD, mangle ("TYPE"), 0, P_UNKNOWN);

	/* Now, stack data and add a call to "TYPE". */
	datap = p_adddata (P_STRING, adjs, strlen (adjs));
	snprintf (buf, sizeof (buf), "R0,'%d", datap->id);
	wp_append ("load", buf, s);
	wp_append ("push", "R0,^D", NULL);
	snprintf (buf, sizeof (buf), "R0,$%d", datap->len);
	wp_append ("load", buf, NULL);
	wp_append ("push", "R0,^D", NULL);
	snprintf (buf, sizeof (buf), "#%d", symp->id);
	wp_append ("call", buf, NULL);
}
static void
f_cstring (char *s)
{
	char		buf[GENERAL_BUF_LEN];
	char		*adjs = adjstr (s);
	struct datarec	*datap;

	/* Add a string to data table, and push its address. */
	datap = p_adddata (P_COUNTEDSTRING, adjs, strlen (adjs));
	snprintf (buf, sizeof (buf), "R0,'%d", datap->id);
	wp_append ("load", buf, s);
	wp_append ("push", "R0,^D", NULL);
}
static void
f_string_literal (char *s)
{
	char		buf[GENERAL_BUF_LEN];
	char		*adjs = adjstr (s);
	struct datarec	*datap;

	/* Add string to data table, and push its address and length. */
	datap = p_adddata (P_STRING, adjs, strlen (adjs));
	snprintf (buf, sizeof (buf), "R0,'%d", datap->id);
	wp_append ("load", buf, s);
	wp_append ("push", "R0,^D", NULL);
	snprintf (buf, sizeof (buf), "R0,$%d", datap->len);
	wp_append ("load", buf, NULL);
	wp_append ("push", "R0,^D", NULL);
}


/*
 * f_integer_literal(),f_float_literal(),f_char_literal()
 *
 * Parse terminal handlers for integer, float, and character literal values.
 */
static void
f_integer_literal (char *s)
{
	char		buf[GENERAL_BUF_LEN];
	union literal {
		int		i;
		unsigned int	u;
	} literal;
	assert (s != NULL);
	assert (sizeof (unsigned int) == sizeof (int));

	/*
	 * Push an integer literal onto the data stack.  Try this as a signed
	 * integer first, then unsigned.
	 */
	errno = 0;
	literal.i = strtol (s, NULL, 0);
	if (errno == ERANGE)
	    {
		if (*s == '-')
			yyerrora ("integer literal '%s' out of range", s);
		else
		    {
			errno = 0;
			literal.u = strtoul (s, NULL, 0);
			if (errno == ERANGE)
				yyerrora ("integer literal '%s' out of range",
									s);
		    }
	    }

	/* Append pcode to load the signed bit pattern of the literal. */
	snprintf (buf, sizeof (buf), "R0,$%d", literal.i);
	wp_append ("load", buf, s);
	wp_append ("push", "R0,^D", NULL);
}
static void
f_float_literal (char *s)
{
	char		buf[GENERAL_BUF_LEN];
	union literal {
		float		f;
		int		i;
	} literal;
	assert (s != NULL);
	assert (sizeof (int) == sizeof (float));

	/*
	 * Push a float literal onto the floating stack.  Convert string to
	 * a float, then push as a signed int bit pattern.
	 */
	if (sscanf (s, "%f", &literal.f) != 1)
		yyerrora ("floating literal '%s' conversion error", s);
	snprintf (buf, sizeof (buf), "R0,$%d", literal.i);
	wp_append ("load", buf, s);
	wp_append ("push", "R0,^F", NULL);
}
static void
f_char_literal (char *s)
{
	char		buf[GENERAL_BUF_LEN];
	char		*sp = skipspace (skipnonspace (s));

	/* Push the ASCII of the first character of s. */
	if (strlen (sp) > 1)
		yywarninga ("warning: char literal '%s' >1 character", sp);
	snprintf (buf, sizeof (buf), "R0,$%d", (int)*sp);
	wp_append ("load", buf, s);
	wp_append ("push", "R0,^D", NULL);
}


/*
 * f_object_address()
 *
 * Parse terminal handler for obtaining a word's address.
 */
static void
f_object_address (char *s)
{
	char		buf[GENERAL_BUF_LEN];
	char		*sp = skipspace (skipnonspace (s));
	struct symrec	*symp = p_findsym (mangle (sp));

	/* Find the object, and push its address. */
	if (symp == NULL)
		symp = p_addsym (P_EXTERNALWORD, mangle (sp), 0, P_UNKNOWN);
	snprintf (buf, sizeof (buf), "R0,#%d", symp->id);
	wp_append ("load", buf, s);
	wp_append ("push", "R0,^D", NULL);
}


/*
 * f_word()
 *
 * Parse terminal handler for a word that isn't any known keyword.
 */
static void
f_word (char *s)
{
	char		buf[GENERAL_BUF_LEN];
	struct symrec	*symp = p_findsym (mangle (s));

	/*
	 * If word is a variable, push its address, if it is a constant,
	 * push its value, and if it is neither treat it as a function call.
	 */
	if (symp == NULL)
		symp = p_addsym (P_EXTERNALWORD, mangle (s), 0, P_UNKNOWN);

	/* Handle the symbol according to its type. */
	switch (symp->type)
	    {
		case P_VARIABLE:
			snprintf (buf, sizeof (buf), "R0,#%d", symp->id);
			wp_append ("load", buf, s);
			wp_append ("push", "R0,^D", NULL);
			break;

		case P_CONSTANT:
			snprintf (buf, sizeof (buf), "R0,@%d", symp->id);
			wp_append ("load", buf, s);
			wp_append ("push", "R0,^D", NULL);
			break;

		default:
			snprintf (buf, sizeof (buf), "#%d", symp->id);
			wp_append ("call", buf, s);
			break;
	    }
}


/*
 * f_if(),f_else(),f_then()
 *
 * Parse terminal handlers for IF...ELSE...THEN constructs.
 */
static void
f_if (char *s)
{
	char		buf[GENERAL_BUF_LEN];
	assert (s != NULL);

	/* Test flag, and jump round if false. */
	wp_append ("pop", "R0,^D", s);
	snprintf (buf, sizeof (buf), "R0,%s", ltoa (labelpush ()));
	wp_append ("jump_z", buf, NULL);
}
static void
f_else (char *s)
{
	int	over = labelpop ();

	/* Jump over, then place label to jump to on if. */
	wp_append ("jump", ltoa (labelpush ()), s);
	wp_append ("label", ltoa (over), NULL);
}
static void
f_then (char *s)
{
	/* Place end-of-ifthenelse label. */
	wp_append ("label", ltoa (labelpop ()), s);
}


/*
 * f_do(),f_query_do(),f_leave(),f_loop(),f_plus_loop(),f_begin(),
 * f_again(),f_until(),f_while(),f_repeat()
 *
 * Parse terminal handlers for all recognized forms of loop construct.
 */
static void
f_do (char *s)
{
	/*
	 * Calculate loop iterations from the top two data stack items,
	 * and push the iterations data onto the return stack.
	 */
	wp_append ("pop", "R1,^D", s);
	wp_append ("pop", "R0,^D", NULL);
	labelpush ();
	wp_append ("push", "R0,^R", NULL);
	wp_append ("subtract", "R1,R0", NULL);
	wp_append ("push", "R1,^R", NULL);
	wp_append ("label", ltoa (labelpush ()), NULL);
	breakpush (labelpush ());
}
static void
f_query_do (char *s)
{
	char	buf[GENERAL_BUF_LEN];

	/* Calculate iterations as for DO, with jump over if equal. */
	wp_append ("pop", "R1,^D", s);
	wp_append ("pop", "R0,^D", NULL);
	snprintf (buf, sizeof (buf), "R0,R1,%s", ltoa (labelpush ()));
	wp_append ("jump_eq", buf, NULL);
	wp_append ("push", "R0,^R", NULL);
	wp_append ("subtract", "R1,R0", NULL);
	wp_append ("push", "R1,^R", NULL);
	wp_append ("label", ltoa (labelpush ()), NULL);
	breakpush (labelpush ());
}
static void
f_leave (char *s)
{
	assert (s != NULL);

	/* Jump out to break label. */
	if (break_index == 0)
		yyerrora ("'%s' invalid outside a loop", s);
	else
		wp_append ("jump", ltoa (breaktos ()), s);
}
static void
f_loop (char *s)
{
	char	buf[GENERAL_BUF_LEN];
	int	endl = labelpop ();

	/* Test loop end, and place labels for break jump. */
	wp_append ("pop", "R0,^R", s);
	wp_append ("add", "R0,$1", NULL);
	wp_append ("push", "R0,^R", NULL);
	snprintf (buf, sizeof (buf), "R0,%s", ltoa (labelpop ()));
	wp_append ("jump_nz", buf, NULL);
	wp_append ("label", ltoa (endl), NULL);
	wp_append ("pop", "R0,^R", NULL);
	wp_append ("pop", "R0,^R", NULL);
	wp_append ("label", ltoa (labelpop ()), NULL);
	breakpop ();
}
static void
f_plus_loop (char *s)
{
	char	buf[GENERAL_BUF_LEN];
	int	endl = labelpop ();

	/* Test loop end, and place labels for break jump. */
	wp_append ("pop", "R1,^D", s);
	wp_append ("pop", "R0,^R", NULL);
	wp_append ("add", "R0,R1", NULL);
	wp_append ("push", "R0,^R", NULL);
	snprintf (buf, sizeof (buf), "R0,%s", ltoa (labelpop ()));
	wp_append ("jump_ge", buf, NULL);
	wp_append ("label", ltoa (endl), NULL);
	wp_append ("pop", "R0,^R", NULL);
	wp_append ("pop", "R0,^R", NULL);
	wp_append ("label", ltoa (labelpop ()), NULL);
	breakpop ();
}
static void
f_begin (char *s)
{
	/* Place loop top label, and reserve break label. */
	wp_append ("label", ltoa (labelpush ()), s);
	breakpush (labelpush ());
}
static void
f_again (char *s)
{
	int	endl = labelpop ();

	/* Jump to loop top, and place break label. */
	wp_append ("jump", ltoa (labelpop ()), s);
	wp_append ("label", ltoa (endl), NULL);
	breakpop ();
}
static void
f_until (char *s)
{
	char	buf[GENERAL_BUF_LEN];
	int	endl = labelpop ();

	/* Test stack top, jump to loop top, and place break label. */
	wp_append ("pop", "R0,^D", s);
	snprintf (buf, sizeof (buf), "R0,%s", ltoa (labelpop ()));
	wp_append ("jump_z", buf, NULL);
	wp_append ("label", ltoa (endl), NULL);
	breakpop ();
}
static void
f_while (char *s)
{
	char	buf[GENERAL_BUF_LEN];
	assert (s != NULL);

	/* Test stack top, and if false jump to loop break. */
	if (break_index == 0)
		yyerrora ("'%s' invalid outside a loop", s);
	else
	    {
		wp_append ("pop", "R0,^D", s);
		snprintf (buf, sizeof (buf), "R0,%s", ltoa (breaktos ()));
		wp_append ("jump_z", buf, NULL);
	    }
}
static void
f_repeat (char *s)
{
	int	endl = labelpop ();

	/* Jump to loop top, and place loop end label. */
	wp_append ("jump", ltoa (labelpop ()), s);
	wp_append ("label", ltoa (endl), NULL);
	breakpop ();
}


/*
 * f_case(),f_of(),f_endof(),f_endcase()
 *
 * Parse terminal handlers for CASE...ENDCASE constructs.
 */
static void
f_case (char *s)
{
	/* No actual code, but push an index for the endcase label. */
	wp_append ("no_op", NULL, s);
	breakpush (labelpush ());
}
static void
f_of (char *s)
{
	char		buf[GENERAL_BUF_LEN];

	/* Test top against top - 1, skip over if different. */
	wp_append ("pop", "R1,^D", s);
	wp_append ("pop", "R0,^D", NULL);
	wp_append ("push", "R0,^D", NULL);
	snprintf (buf, sizeof (buf), "R0,R1,%s", ltoa (labelpush ()));
	wp_append ("jump_ne", buf, NULL);
	wp_append ("pop", "R0,^D", NULL);
}
static void
f_endof (char *s)
{
	/* Jump to end of case, then place skip label. */
	wp_append ("jump", ltoa (breaktos ()), s);
	wp_append ("label", ltoa (labelpop ()), NULL);
}
static void
f_endcase (char *s)
{
	/* Pop case value, place end of case label. */
	wp_append ("pop", "R0,^D", s);
	wp_append ("label", ltoa (labelpop ()), NULL);
	breakpop ();
}


/*
 * f_codeword(),f_code(),f_endcode()
 *
 * Special parse terminal handlers for pass-through assembler (is there
 * some easier way than this?).
 */
static void
f_codeword (char *s)
{
	char		buf[GENERAL_BUF_LEN];
	struct symrec	*symp = p_findsym (mangle (s));

	/*
	 * Add word to the symbol table if not yet defined, or alter its
	 * definition if it is.  If known, but undefined, change it into
	 * a real codeword.
	 */
	if (symp != NULL)
	    {
		if (symp->type != P_EXTERNALWORD)
			yyerrora ("multiple symbol definition '%s'", s);
		else
			p_chgsym (symp, P_CODEWORD, 1, P_CELL);
	    }
	else
		/* Add a new code word symbol. */
		symp = p_addsym (P_CODEWORD, mangle (s), 1, P_CELL);

	/* Set this to be the current definition, and start pcode define. */
	currdef = symp;
	snprintf (buf, sizeof (buf), "#%d", symp->id);
	wp_append ("define", buf, s);
}
static void
f_code (char *s)
{
	char	*sp;
	assert (s != NULL);

	/*
	 * Add the code line to the generated code trail.  First, strip
	 * [;|#]... comments; it's very awkward to get flex to do it.
	 */
	for (sp = s; *sp != '\0'; sp++)
	    {
		/*
		 * If the current character is '"', scan forward to the
		 * next '"', or the end of line, to avoid finding ';' or
		 * '#' in strings.
		 */
		if (*sp == '"')
			for (sp++; *sp != '\0' && *sp != '"'; sp++)
				if (*sp == '\\')
					sp++;

		/* Check for a comment indicator; one of ';' or '#'. */
		if (strchr ("#;", *sp) != NULL)
		    {
			/* Terminate the string here. */
			*sp = '\0';

			/*
			 * For neatness, back up through any whitespace we
			 * find ahead of the comment indicator we just removed.
			 */
			for (sp--; sp >= s && isspace (*sp); sp--)
				*sp = '\0';
		    }
	    }

	/* Write the code verbatim, if there is anything left. */
	if (strlen (s) > 0)
		wp_append ("asm", s, NULL);
}
static void
f_end_code (char *s)
{
	/* End definition. */
	wp_append ("enddef", NULL, s);
	currdef = NULL;
}


/*
 * add_main_symbol()
 *
 * Helper function, adds the given string as a symbol in the main module.
 * The code generator expects to find these symbols in order to copy
 * information from the stack on process startup to them.
 */
static void
add_main_symbol (char *name)
{
	struct symrec	*symp = p_findsym (mangle (name));

	/* If the symbol is not new or undefined, it should be. */
	if (symp != NULL)
	    {
		if (symp->type != P_EXTERNALWORD)
			yyerrora ("'%s' defined in main module", name);
		else
			p_chgsym (symp, P_VARIABLE, 1, P_CELL);
	    }
	else
		/* Add the symbol, even though its not used in this module. */
		p_addsym (P_VARIABLE, mangle (name), 1, P_CELL);
}


/*
 * yy_parse_ifile()
 *
 * Parse file entry function.
 */
int yy_parse_ifile (void) {
	int	yyparse_status = 0;

	/* Clear down states, counts, and sequences from any prior parse. */
	yy_linenum	= 1;
	yy_errcount	= 0;
	yy_warncount	= 0;

	/* Clean label sequence and stack. */
	label_index	= 0;
	label_seq	= 1;

	/* Clear definition end labels. */
	end_label	= 0;
	need_main_end	= FALSE;
	has_main	= FALSE;

	/* Clear any pending definitions and breaks. */
	currdef		= NULL;
	break_index	= 0;

	/* Set the input file for the lexical analyser. */
	yyrestart (get_ifstream ());

	/* Run yyparse and note the result. */
	in_yyparse	= TRUE;
	yyparse_status	= yyparse ();
	in_yyparse	= FALSE;

	/* If no errors, check that both our parse stacks are empty. */
	if (yy_errcount == 0)
	    {
		if (label_index > 0)
			yyerror ("internal error: parse stack not emptied");
		if (break_index > 0)
			yyerror ("internal error: loop stack not emptied");
	    }

	/* If this is a main module, add the main module arg/env symbols. */
	if (has_main)
	    {
		add_main_symbol ("_argc");
		add_main_symbol ("_argv");
		add_main_symbol ("_envp");
	    }

	/*
	 * If this module needs a main end, append it to the pstore.  Use
	 * yy_linenum - 1 because since the parse is actually over, the
	 * line number count is in fact one past the last line.
	 */
	if (need_main_end)
		p_append ("label", ltoa (end_label), NULL, yy_linenum - 1);

	/*
	 * If there's nothing in the data, symbol, or pcode areas of the
	 * pstore, and we've not mentioned any errors, then we should
	 * probably warn that the whole file contained nothing as far as
	 * we could tell.
	 */
	if (yy_errcount == 0)
	    {
		if (p_datastart() == NULL
				&& p_symstart() == NULL
				&& p_codestart () == NULL)
			yywarning (
				"warning: file contained no symbols or code");
	    }

	/* Return the result of running yyparse. */
	return yyparse_status;
}
