/*
 * VNPForth - Compiled native Forth for IA32 Linux systems
 * Copyright (C) 2001  Simon Baldwin (simonb@caldera.com)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <assert.h>
#include <string.h>
#include <limits.h>
#include <stdio.h>

#include "ident.h"

/* Definitions. */
#define	IDENT_MAX		4096
#define	IDENT_HIGH_WATER	(IDENT_MAX - 64)

/* True/False macros. */
#define FALSE			(0)
#define TRUE			(!FALSE)

/* Character entity lookup table. */
static const struct entrec {
	const char		character;
	const char		*name;
	const int		namelen;
} enttable[] = {
	{ '`', "Backtick", 8 },		{ '~', "Tilde", 5 },
	{ '!', "Store", 5 },		{ '@', "Fetch", 5 },
	{ '#', "Hash", 4 },		{ '$', "Dollar", 6 },
	{ '%', "Percent", 7 },		{ '^', "Caret", 5 },
	{ '&', "Amp", 3 },		{ '*', "Star", 4 },
	{ '(', "Paren", 5 },		{ ')', "Closeparen", 10 },
	{ '-', "Minus", 5 },		{ '=', "Equals", 6 },
	{ '+', "Plus", 4 },		{ '[', "Leftbracket", 11 },
	{ ']', "Rightbracket", 12 },	{ '{', "Leftbrace", 9 },
	{ '}', "Rightbrace", 10 },	{ '|', "Or", 2 },
	{ ';', "Semicolon", 9 },	{ ':', "Colon", 5 },
	{ '"', "Quote", 5 },		{ '\'', "Tick", 4 },
	{ ',', "Comma", 5 },		{ '.', "Dot", 3 },
	{ '/', "Slash", 5 },		{ '<', "Less", 4 },
	{ '>', "Greater", 7 },		{ '?', "Question", 8 },
	{ '\\', "Backslash", 9 },
	{ '\0', NULL, 0 }
};

/* Hexadecimal characters array. */
static const char	*hex_chars = "0123456789ABCDEF";


/*
 * lookup()
 *
 * Lookup the string (entity) equivalent for a particular character.
 * Return NULL if there is no string equivalent suitable.
 */
static const struct entrec *
lookup (char character)
{
	static const struct entrec
			*entlookup[CHAR_MAX];	/* Direct entry lookup array */
	static int	lookupinit = FALSE;	/* Array initialized flag */
	const struct entrec
			*entp;			/* Pointer through enttable */

	/* If the faster lookup table is not initialized, do it now. */
	if (!lookupinit)
	    {
		/*
		 * For each table entry, assign an array element that points
		 * to the table entry itself.  Unassigned fast lookup array
		 * elements get to be NULL courtesy of static data initial-
		 * ization rules.
		 */
		for (entp = enttable; entp->name != NULL; entp++)
		    {
			assert (entp->character >= 0);
			assert (entp->namelen == (int)(strlen (entp->name)));
			entlookup[(int) (entp->character)] = entp;
		    }

		/* Note that the table is built. */
		lookupinit = TRUE;
	    }

	/*
	 * Now, simply return the pointer from the entlookup table at the
	 * index of 'character', or NULL if there is nothing there.
	 */
	return entlookup[(int) character];
}


/*
 * mangle()
 *
 * Convert an identifier into something that C/asm can handle.
 */
char *
mangle (const char *name)
{
	static char	buf[IDENT_MAX];	/* Return buffer, static */
	const char	*sp;		/* Pointer through name string */
	char		*bufp;		/* Pointer through return buffer */
	assert (name != NULL);

	/* Begin with a clean buffer; all characters set to NUL. */
	memset (buf, '\0', sizeof (buf));

	/*
	 * Handle each and every character in the name, but stop if the
	 * output buffer high water mark is reached.
	 */
	for (sp = name, bufp = buf;
			*sp != '\0' && bufp < buf + IDENT_HIGH_WATER; sp++)
	    {
		char c = *sp;

		/* If this is a normal character, probably copy verbatim... */
		if (c == '_'	|| (c >= 'A' && c <= 'Z')
				|| (c >= '0' && c <= '9')
				|| (c >= 'a' && c <= 'z'))
		    {
			/* ...except if the first character is a digit. */
			if (sp == name && c >= '0' && c <= '9')
				*bufp++ = 'z';
			*bufp++ = c;
		    }
		else
		    {
			const struct entrec	*entp = lookup (c);

			/* Suppress the separator underscore if at the start. */
			if (sp != name)
				*bufp++ = '_';

			/* If an entity name exists for the char, use it. */
			if (entp != NULL)
			    {
				memcpy (bufp, entp->name, entp->namelen);
				bufp += entp->namelen;
			    }
			else
			    {
				/* There is nothing for it but to insert hex. */
				*bufp++ = 'x';
				*bufp++ = hex_chars[(c & 0xF0) >> 4];
				*bufp++ = hex_chars[(c & 0x0F)     ];
			    }
		    }
	    }

	/* Done - return buffer. */
	return buf;
}
