/*
 * VNPForth - Compiled native Forth for IA32 Linux systems
 * Copyright (C) 2001  Simon Baldwin (simonb@caldera.com)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <assert.h>
#include <time.h>
#include <string.h>
#include <stdio.h>
#include <malloc.h>
#include <errno.h>

#include "ifile.h"

#include "pstore.h"

/* Memory structure magic numbers, for identification and reassurance. */
#define	SYMREC_MAGIC	0x899A71FB
#define	DATAREC_MAGIC	0x3BF0BAE6
#define	PCODEREC_MAGIC	0x716D56FC

/* Definitions for the hash algorithm. */
#define NBUCKET_PRIME	211

/* Growable array increment, in array elements. */
#define	IDINCREMENT	32

/* Symbol and data table (list) heads and tails, hash heads. */
static struct datarec	*datalist	= NULL;
static struct datarec	*datatail	= NULL;
static struct symrec	*symlist	= NULL;
static struct symrec	*symtail	= NULL;
static struct symrec	*symhashlist[NBUCKET_PRIME];

/* Data structure that can represent either a symbol or a data record. */
union record {
	struct symrec	symrec;
	struct datarec	datarec;
};

/* Growable array of symbols for direct access by id. */
static union record	**idarray	= NULL;
static int		iddimension	= 0;
static int		idsize		= 0;

/* Pcode list head and tail. */
static struct pcoderec	*pcodelist	= NULL;
static struct pcoderec	*pcodetail	= NULL;


/*
 * wmalloc()
 *
 * Error-handling wrapper round malloc.
 */
static void *
wmalloc (int size)
{
	void *result = malloc (size);

	if (result == NULL)
	    {
		fprintf (stderr, "Out of memory\n");
		exit (1);
	    }
	return result;
}


/*
 * copy_string()
 *
 * Malloc space for a string, copy the string, and return the address.
 */
static char *
copy_string (const char *s)
{
	char	*result	= wmalloc (strlen (s) + 1);
						/* Result string to return */

	strcpy (result, s);
	return result;
}


/*
 * new_id()
 *
 * Return a new symbol or data record id, and store the record pointer
 * in the id array, at offset [id].
 */
static int
new_id (union record *rec)
{
	/* Grow the id array if it is currently full. */
	if (idsize >= iddimension)
	    {
		iddimension += IDINCREMENT;
		idarray = realloc (idarray, iddimension * sizeof (idarray[0]));
		if (idarray == NULL)
		    {
			fprintf (stderr, "Out of memory\n");
			exit (1);
		    }
	    }

	/* Add the new record to the array, and return the index of it. */
	idarray[idsize++] = rec;
	return idsize - 1;
}


/*
 * hash_sym()
 *
 * Hash a symbol name into its hash code - hashpjw algorithm.
 */
static int
hash_sym (const char *s)
{
	const char	*sp;			/* Input string pointer */
	unsigned int	h = 0, g;		/* Hash, and aux variable */

	for (sp = s; *sp != '\0'; sp++)
	    {
		h = (h << 4) + (*sp);
		if ((g = (h & 0xF0000000)) != 0)
		    {
			h = h ^ (g >> 24);
			h = h ^ g;
		    }
	    }
	return h % NBUCKET_PRIME;
}


/*
 * p_addsym()
 *
 * Add a symbol to the symbol table.
 */
struct symrec *
p_addsym (enum symtype type, const char *symname, int len, enum elunits units)
{
	struct symrec	*symp;			/* New symbol table entry */

	/* Create the new entry. */
	symp = wmalloc (sizeof (struct symrec));
	symp->magic	= SYMREC_MAGIC;
	symp->id	= new_id ((union record *) symp);
	symp->type	= type;
	symp->name	= copy_string (symname);
	symp->len	= len;
	symp->units	= units;
	symp->next	= NULL;
	symp->hash	= hash_sym (symname);
	symp->next_hash	= NULL;

	/* Add the new entry to the main list tail. */
	if (symtail == NULL)
		symlist = symtail = symp;
	else
		symtail->next = symtail = symp;

	/* Add the new entry to the hash list head. */
	if (symhashlist[symp->hash] == NULL)
		symhashlist[symp->hash] = symp;
	else
	    {
		symp->next_hash = symhashlist[symp->hash];
		symhashlist[symp->hash] = symp;
	    }
	return symp;
}


/*
 * p_chgsym()
 *
 * Change a symbol's type, length, and units in the symbol table.
 */
void
p_chgsym (struct symrec *symp, enum symtype type, int len, enum elunits units)
{
	assert (symp != NULL && symp->magic == SYMREC_MAGIC);

	/* Modify the entry. */
	symp->type	= type;
	symp->len	= len;
	symp->units	= units;
}


/*
 * p_findsym()
 *
 * Find an entry in the symbol table, by name.
 */
struct symrec *
p_findsym (const char *symname)
{
	struct symrec	*symp;			/* Return symbol table entry */
	int		hash = hash_sym (symname);
						/* Hash of symbol requested */

	/* If no table return null. */
	if (symhashlist[hash] == NULL)
		return NULL;
	else
	    {
		/* Check each hash list entry for the name. */
		for (symp = symhashlist[hash]; symp != NULL;
						symp = symp->next_hash)
			if (strcmp (symp->name, symname) == 0)
				return symp;
	    }

	/* Not found, so return null. */
	return NULL;
}


/*
 * p_adddata()
 *
 * Add a data element to the data table.
 */
struct datarec *
p_adddata (enum datatype type, void *data, int len)
{
	struct datarec		*datap;		/* New data table entry */

	/* Create the new entry. */
	datap = wmalloc (sizeof (struct datarec));
	datap->magic	= DATAREC_MAGIC;
	datap->id	= new_id ((union record *) datap);
	datap->type	= type;
	datap->data	= wmalloc (len);
	memcpy (datap->data, data, len);
	datap->len	= len;
	datap->next	= NULL;

	/* Add the new entry to the list tail. */
	if (datatail == NULL)
		datalist = datatail = datap;
	else
		datatail->next = datatail = datap;
	return datap;
}


/*
 * p_append()
 *
 * Add a pcode entry to the store.
 */
void
p_append (const char *opcode, const char *operands, const char *source_token,
		int source_linenum)
{
	struct pcoderec		*pcodep;	/* New pcode list entry */

	/* Create and fill out the new entry. */
	pcodep = wmalloc (sizeof (struct pcoderec));
	pcodep->magic	= PCODEREC_MAGIC;
	pcodep->opcode	= copy_string (opcode);
	if (operands != NULL)
		pcodep->operands = copy_string (operands);
	else
		pcodep->operands = NULL;
	if (source_token != NULL)
		pcodep->source_token = copy_string (source_token);
	else
		pcodep->source_token = NULL;
	pcodep->linenum	= source_linenum;
	pcodep->next	= NULL;

	/* Add this to the end of the store. */
	if (pcodelist == NULL)
		pcodelist = pcodetail = pcodep;
	else
		pcodetail->next = pcodetail = pcodep;
}


/*
 * p_replace()
 *
 * Replace elements of a pcode entry already in the store.
 */
void
p_replace (struct pcoderec *pcodep, const char *opcode, const char *operands)
{
	assert (pcodep != NULL && pcodep->magic == PCODEREC_MAGIC);

	/* Remove the current record details. */
	free (pcodep->opcode);
	if (pcodep->operands != NULL)
		free (pcodep->operands);

	/* Now add the new details into the pcode entry. */
	pcodep->opcode = copy_string (opcode);
	if (operands != NULL)
		pcodep->operands = copy_string (operands);
	else
		pcodep->operands = NULL;
}


/*
 * p_lookupsym()
 * p_lookupdata()
 *
 * Find symbol and data table entries by id.  In practice, this means taking
 * them out of the id array, and checking the type before returning.
 */
struct symrec *
p_lookupsym (int symid)
{
	struct symrec	*symp;			/* Return symbol table entry */

	/* Return the entry from the id array if in range. */
	if (symid >=0 && symid < idsize)
	    {
		symp = (struct symrec *) idarray[symid];
		if (symp->magic == SYMREC_MAGIC)
		    {
			assert (symp->id == symid);
			return symp;
		    }
	    }
	return NULL;
}
struct datarec *
p_lookupdata (int dataid)
{
	struct datarec	*datap;			/* Return data table entry */

	/* Return the entry from the id array if in range. */
	if (dataid >=0 && dataid < idsize)
	    {
		datap = (struct datarec *) idarray[dataid];
		if (datap->magic == DATAREC_MAGIC)
		    {
			assert (datap->id == dataid);
			return datap;
		    }
	    }
	return NULL;
}


/*
 * p_codestart()
 * p_codenext()
 *
 * Iterator to return pcode entries, start, and sequenced.
 */
struct pcoderec *
p_codestart (void)
{
	return pcodelist;
}
struct pcoderec *
p_codenext (struct pcoderec *current)
{
	assert (current != NULL && current->magic == PCODEREC_MAGIC);
	return current->next;
}


/*
 * p_symstart()
 * p_symnext()
 * p_datastart()
 * p_datanext()
 *
 * Iterators, to return symbol and data table entries, start, and sequenced.
 */
struct symrec *
p_symstart (void)
{
	return symlist;
}
struct symrec *
p_symnext (struct symrec *current)
{
	assert (current != NULL && current->magic == SYMREC_MAGIC);
	return current->next;
}
struct datarec *
p_datastart (void)
{
	return datalist;
}
struct datarec *
p_datanext (struct datarec *current)
{
	assert (current != NULL && current->magic == DATAREC_MAGIC);
	return current->next;
}


/*
 * p_empty()
 *
 * Clear and free the complete store.
 */
void
p_empty (void)
{
	struct symrec	*symp, *nextsymp;	/* Pointers to symbol entries */
	struct datarec	*datap, *nextdatap;	/* Pointers to data entries */
	struct pcoderec	*pcodep, *nextpcodep;	/* Pointers to pcode entries */
	int		index;			/* Index for clearing hashes */

	/* Destroy the symbol table list. */
	for (symp = symlist; symp != NULL; symp = nextsymp)
	    {
		nextsymp = symp->next;
		symp->magic = 0;
		free (symp->name);
		free (symp);
	    }
	symlist = symtail = NULL;

	/* Reset the hash list heads to all NULLs. */
	for (index = 0; index < NBUCKET_PRIME; index++)
		symhashlist[index] = NULL;

	/* Destroy the data table list. */
	for (datap = datalist; datap != NULL; datap = nextdatap)
	    {
		nextdatap = datap->next;
		datap->magic = 0;
		free (datap->data);
		free (datap);
	    }
	datalist = datatail = NULL;

	/* Destroy the growable id array. */
	iddimension = idsize = 0;
	free (idarray);
	idarray = NULL;

	/* Destroy the complete pcode list. */
	for (pcodep = pcodelist; pcodep != NULL; pcodep = nextpcodep)
	    {
		nextpcodep = pcodep->next;
		pcodep->magic = 0;
		free (pcodep->opcode);
		if (pcodep->operands != NULL)
			free (pcodep->operands);
		if (pcodep->source_token != NULL)
			free (pcodep->source_token);
		free (pcodep);
	    }
	pcodelist = pcodetail = NULL;
}


/*
 * p_printout()
 *
 * Dump the complete pstore to a stream in human-readable format.
 */
void
p_printout (FILE * stream)
{
	const char *symcunits[]	= { "CELL", "CHAR", "" };
	const char *symctype[]	= { "VARIABLE", "CONSTANT", "EXTERNAL", "WORD",
	       			    "CODEWORD" };
	const char *datactype[]	= { "DATA", "STRING", "CSTRING" };
	time_t	now;				/* For time/localtime */

	fprintf (stream, "Forth 1.1, %s %s, Linux\n\n", __DATE__, __TIME__);
	fprintf (stream, "File:     %s\n", get_ifname ());
	time (&now);
	fprintf (stream, "Compiled: %s\n\n", asctime (localtime (&now)));

	fprintf (stream, "Data table:\n");
	if (datalist == NULL)
		fprintf (stream, "(empty)\n");
	else
	    {
		struct datarec *datap;
		fprintf (stream, "%-8s %-8s %-6s %s\n",
			 "RecordID", "Type", "Length", "Data");
		fprintf (stream, "%-8s %-8s %-6s %s\n",
			 "--------", "----", "------", "----");
		for (datap = datalist; datap != NULL; datap = datap->next)
		    {
			fprintf (stream, "[%6u] %-8s %6d ", datap->id,
				 datactype[datap->type], datap->len);
			if (datap->type == P_STRING
					|| datap->type == P_COUNTEDSTRING)
			    {
				int i;
				fprintf (stream, "\"");
				for (i = 0; i < datap->len; i++)
					fprintf (stream, "%c",
						 *((char *) (datap->data) + i));
				fprintf (stream, "\"\n");
			    }
			else
			    {
				int i;
				fprintf (stream, "<");
				for (i = 0; i < datap->len; i++)
					fprintf (stream, "%02X ",
						 *((char *) (datap->data) + i));
				fprintf (stream, ">\n");
			    }
		}
	}
	fprintf (stream, "\n");

	fprintf (stream, "Symbol table:\n");
	if (symlist == NULL)
		fprintf (stream, "(empty)\n");
	else
	    {
		struct symrec *symp;
		fprintf (stream, "%-8s %-8s %-6s %-5s %-4s %s\n",
			 "RecordID", "Type", "Length", "Units", "Hash",
			 "Name");
		fprintf (stream, "%-8s %-8s %-6s %-5s %-4s %s\n",
			 "--------", "----", "------", "-----", "----",
			 "----");
		for (symp = symlist; symp != NULL; symp = symp->next)
		    {
			fprintf (stream, "[%6u] %-8s %6d %-5s %4d %s\n",
				symp->id, symctype[symp->type], symp->len,
				symcunits[symp->units], symp->hash, symp->name);
		    }
	    }
	fprintf (stream, "\n");

	fprintf (stream, "Intermediate code:\n");
	if (pcodelist == NULL)
		fprintf (stream, "(empty)\n");
	else
	    {
		struct pcoderec *pcodep;
		int i = 0;
		fprintf (stream, "%-8s %-6s %-8s %-30s %s\n",
			 "Line Num", "Seqnce", "Opcode", "Operands", "Source");
		fprintf (stream, "%-8s %-6s %-8s %-30s %s\n",
			 "--------", "------", "------", "--------", "------");
		for (pcodep = pcodelist; pcodep != NULL; pcodep = pcodep->next)
		    {
			fprintf (stream, "[%6d] %6d %-8s ", pcodep->linenum,
				 i++, pcodep->opcode);
			if (pcodep->source_token != NULL)
			    {
				if (pcodep->operands != NULL)
					fprintf (stream, "%-30s ",
						 pcodep->operands);
				else
					fprintf (stream, "%-30s ", "");
				fprintf (stream, "%s", pcodep->source_token);
			    }
			else
			    {
				if (pcodep->operands != NULL)
					fprintf (stream, "%s",
						 pcodep->operands);
			    }
			fprintf (stream, "\n");
		    }
	    }
	fprintf (stream, "\n");
}
